@@ -4,7 +4,8 @@
 #include "utils.h"
 #include "H5Zlzo.h"                    /* Import FILTER_LZO */
 #include "H5Zbzip2.h"                  /* Import FILTER_BZIP2 */
-#include "../blosc/blosc_filter.h"     /* Import FILTER_BLOSC */
+#include "blosc_filter.h"              /* Import FILTER_BLOSC */
+#include "blosc2_filter.h"             /* Import FILTER_BLOSC2 */
 
 #include <string.h>
 #include <stdlib.h>
@@ -27,11 +28,12 @@
  * Date: March 19, 2001
  *
  * Comments: Modified by F. Alted. November 07, 2003
+ *           Modified by A. Cobb. August 21, 2017 (track_times)
  *
  *-------------------------------------------------------------------------
  */
 
-herr_t H5ARRAYmake( hid_t loc_id,
+hid_t H5ARRAYmake(  hid_t loc_id,
                     const char *dset_name,
                     const char *obversion,
                     const int rank,
@@ -44,13 +46,16 @@ herr_t H5ARRAYmake( hid_t loc_id,
                     char  *complib,
                     int   shuffle,
                     int   fletcher32,
+		    hbool_t track_times,
                     const void *data)
 {
 
  hid_t   dataset_id, space_id;
  hsize_t *maxdims = NULL;
  hid_t   plist_id = 0;
- unsigned int cd_values[6];
+ unsigned int cd_values[7];
+ int     blosc_compcode;
+ char    *blosc_compname = NULL;
  int     chunked = 0;
  int     i;
 
@@ -77,9 +82,15 @@ herr_t H5ARRAYmake( hid_t loc_id,
  if ( (space_id = H5Screate_simple( rank, dims, maxdims )) < 0 )
    return -1;
 
+ /* Create dataset creation property list with default values */
+ plist_id = H5Pcreate (H5P_DATASET_CREATE);
+
+ /* Enable or disable recording dataset times */
+ if ( H5Pset_obj_track_times( plist_id, track_times ) < 0 )
+   return -1;
+
  if (chunked) {
    /* Modify dataset creation properties, i.e. enable chunking  */
-   plist_id = H5Pcreate (H5P_DATASET_CREATE);
    if ( H5Pset_chunk ( plist_id, rank, dims_chunk ) < 0 )
      return -1;
 
@@ -102,8 +113,8 @@ herr_t H5ARRAYmake( hid_t loc_id,
      if ( H5Pset_fletcher32( plist_id) < 0 )
        return -1;
    }
-   /* Then shuffle (not if blosc is activated) */
-   if ((shuffle) && (strcmp(complib, "blosc") != 0)) {
+   /* Then shuffle (blosc shuffles inplace) */
+   if ((shuffle && compress) && (strncmp(complib, "blosc", 5) != 0)) {
      if ( H5Pset_shuffle( plist_id) < 0 )
        return -1;
    }
@@ -121,6 +132,23 @@ herr_t H5ARRAYmake( hid_t loc_id,
        if ( H5Pset_deflate( plist_id, compress) < 0 )
          return -1;
      }
+     /* The Blosc2 compressor does accept parameters */
+     else if (strcmp(complib, "blosc2") == 0) {
+       cd_values[4] = compress;
+       cd_values[5] = shuffle;
+       if ( H5Pset_filter( plist_id, FILTER_BLOSC2, H5Z_FLAG_OPTIONAL, 6, cd_values) < 0 )
+         return -1;
+      }
+      /* The Blosc2 compressor can use other compressors */
+     else if (strncmp(complib, "blosc2:", 7) == 0) {
+       cd_values[4] = compress;
+       cd_values[5] = shuffle;
+       blosc_compname = complib + 7;
+       blosc_compcode = blosc2_compname_to_compcode(blosc_compname);
+       cd_values[6] = blosc_compcode;
+       if ( H5Pset_filter( plist_id, FILTER_BLOSC2, H5Z_FLAG_OPTIONAL, 7, cd_values) < 0 )
+         return -1;
+     }
      /* The Blosc compressor does accept parameters */
      else if (strcmp(complib, "blosc") == 0) {
        cd_values[4] = compress;
@@ -128,6 +156,16 @@ herr_t H5ARRAYmake( hid_t loc_id,
        if ( H5Pset_filter( plist_id, FILTER_BLOSC, H5Z_FLAG_OPTIONAL, 6, cd_values) < 0 )
          return -1;
      }
+     /* The Blosc compressor can use other compressors */
+     else if (strncmp(complib, "blosc:", 6) == 0) {
+       cd_values[4] = compress;
+       cd_values[5] = shuffle;
+       blosc_compname = complib + 6;
+       blosc_compcode = blosc_compname_to_compcode(blosc_compname);
+       cd_values[6] = blosc_compcode;
+       if ( H5Pset_filter( plist_id, FILTER_BLOSC, H5Z_FLAG_OPTIONAL, 7, cd_values) < 0 )
+         return -1;
+     }
      /* The LZO compressor does accept parameters */
      else if (strcmp(complib, "lzo") == 0) {
        if ( H5Pset_filter( plist_id, FILTER_LZO, H5Z_FLAG_OPTIONAL, 3, cd_values) < 0 )
@@ -153,7 +191,7 @@ herr_t H5ARRAYmake( hid_t loc_id,
  else {         /* Not chunked case */
    /* Create the dataset. */
    if ((dataset_id = H5Dcreate(loc_id, dset_name, type_id, space_id,
-                               H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT )) < 0 )
+                               H5P_DEFAULT, plist_id, H5P_DEFAULT )) < 0 )
      goto out;
  }
 
