@@ -1,30 +1,17 @@
-# -*- coding: utf-8 -*-
-
-########################################################################
-#
-# License: BSD
-# Created: October 14, 2002
-# Author: Francesc Alted - faltet@pytables.com
-#
-# $Id$
-#
-########################################################################
-
 """Here is defined the Leaf class."""
 
 import warnings
 import math
 
-import numpy
+import numpy as np
+import cpuinfo
 
-from tables.flavor import (check_flavor, internal_flavor,
-                           alias_map as flavor_alias_map)
-from tables.node import Node
-from tables.filters import Filters
-from tables.utils import byteorders, lazyattr, SizeType
-from tables.exceptions import PerformanceWarning
-from tables import utilsextension
-from tables._past import previous_api
+from .flavor import (check_flavor, internal_flavor, toarray,
+                     alias_map as flavor_alias_map)
+from .node import Node
+from .filters import Filters
+from .utils import byteorders, lazyattr, SizeType
+from .exceptions import PerformanceWarning
 
 
 def csformula(expected_mb):
@@ -33,7 +20,7 @@ def csformula(expected_mb):
     # For a basesize of 8 KB, this will return:
     # 8 KB for datasets <= 1 MB
     # 1 MB for datasets >= 10 TB
-    basesize = 8 * 1024   # 8 KB is a good minimum
+    basesize = 8 * 1024  # 8 KB is a good minimum
     return basesize * int(2**math.log10(expected_mb))
 
 
@@ -68,8 +55,8 @@ def calc_chunksize(expected_mb):
     zone = int(math.log10(expected_mb))
     expected_mb = 10**zone
     chunksize = csformula(expected_mb)
-    return chunksize * 8     # XXX: Multiply by 8 seems optimal for
-                           # sequential access
+    # XXX: Multiply by 8 seems optimal for sequential access
+    return chunksize * 8
 
 
 class Leaf(Node):
@@ -123,11 +110,6 @@ class Leaf(Node):
 
     """
 
-    # Properties
-    # ~~~~~~~~~~
-
-    # Node property aliases
-    # `````````````````````
     # These are a little hard to override, but so are properties.
     attrs = Node._v_attrs
     """The associated AttributeSet instance - see :ref:`AttributeSetClassDescr`
@@ -136,41 +118,39 @@ class Leaf(Node):
     """A description for this node
     (This is an easier-to-write alias of :attr:`Node._v_title`)."""
 
-    # Read-only node property aliases
-    # ```````````````````````````````
-    name = property(
-        lambda self: self._v_name, None, None,
-        """The name of this node in its parent group
-        (This is an easier-to-write alias of :attr:`Node._v_name`).""")
+    @property
+    def name(self):
+        """The name of this node in its parent group (This is an
+        easier-to-write alias of :attr:`Node._v_name`)."""
+        return self._v_name
 
-    chunkshape = property(
-        lambda self: getattr(self, '_v_chunkshape', None), None, None,
+    @property
+    def chunkshape(self):
         """The HDF5 chunk size for chunked leaves (a tuple).
 
         This is read-only because you cannot change the chunk size of a
         leaf once it has been created.
-        """)
+        """
+        return getattr(self, '_v_chunkshape', None)
 
-    object_id = property(
-        lambda self: self._v_objectid, None, None,
+    @property
+    def object_id(self):
         """A node identifier, which may change from run to run.
         (This is an easier-to-write alias of :attr:`Node._v_objectid`).
 
         .. versionchanged:: 3.0
            The *objectID* property has been renamed into *object_id*.
 
-        """)
+        """
+        return self._v_objectid
 
-    objectID = previous_api(object_id)
-
-    ndim = property(
-        lambda self: len(self.shape), None, None,
+    @property
+    def ndim(self):
         """The number of dimensions of the leaf data.
 
-        .. versionadded: 2.4""")
+        .. versionadded: 2.4"""
+        return len(self.shape)
 
-    # Lazy read-only attributes
-    # `````````````````````````
     @lazyattr
     def filters(self):
         """Filter properties for this leaf.
@@ -183,32 +163,34 @@ class Leaf(Node):
 
         return Filters._from_leaf(self)
 
-    # Other properties
-    # ````````````````
-    def _getmaindim(self):
-        if self.extdim < 0:
-            return 0  # choose the first dimension
-        return self.extdim
+    @property
+    def track_times(self):
+        """Whether timestamps for the leaf are recorded
 
-    maindim = property(
-        _getmaindim, None, None,
+        If the leaf is not a dataset, this will fail with HDF5ExtError.
+
+        The track times dataset creation property does not seem to
+        survive closing and reopening as of HDF5 1.8.17.  Currently,
+        it may be more accurate to test whether the ctime for the
+        dataset is 0:
+        track_times = (leaf._get_obj_timestamps().ctime == 0)
+        """
+        return self._get_obj_track_times()
+
+    @property
+    def maindim(self):
         """The dimension along which iterators work.
 
         Its value is 0 (i.e. the first dimension) when the dataset is not
         extendable, and self.extdim (where available) for extendable ones.
-        """)
+        """
 
-    def _setflavor(self, flavor):
-        self._v_file._check_writable()
-        check_flavor(flavor)
-        self._v_attrs.FLAVOR = self._flavor = flavor  # logs the change
+        if self.extdim < 0:
+            return 0  # choose the first dimension
+        return self.extdim
 
-    def _delflavor(self):
-        del self._v_attrs.FLAVOR
-        self._flavor = internal_flavor
-
-    flavor = property(
-        lambda self: self._flavor, _setflavor, _delflavor,
+    @property
+    def flavor(self):
         """The type of data object read from this leaf.
 
         It can be any of 'numpy' or 'python'.
@@ -216,21 +198,35 @@ class Leaf(Node):
         You can (and are encouraged to) use this property to get, set
         and delete the FLAVOR HDF5 attribute of the leaf. When the leaf
         has no such attribute, the default flavor is used..
-        """)
+        """
 
-    size_on_disk = property(lambda self: self._get_storage_size(), None, None,
-                            """
+        return self._flavor
+
+    @flavor.setter
+    def flavor(self, flavor):
+        self._v_file._check_writable()
+        check_flavor(flavor)
+        self._v_attrs.FLAVOR = self._flavor = flavor  # logs the change
+
+    @flavor.deleter
+    def flavor(self):
+        del self._v_attrs.FLAVOR
+        self._flavor = internal_flavor
+
+    @property
+    def size_on_disk(self):
+        """
         The size of this leaf's data in bytes as it is stored on disk.  If the
         data is compressed, this shows the compressed size.  In the case of
         uncompressed, chunked data, this may be slightly larger than the amount
         of data, due to partially filled chunks.
-        """)
+        """
+        return self._get_storage_size()
 
-    # Special methods
-    # ~~~~~~~~~~~~~~~
     def __init__(self, parentnode, name,
                  new=False, filters=None,
-                 byteorder=None, _log=True):
+                 byteorder=None, _log=True,
+                 track_times=True):
         self._v_new = new
         """Is this the first time the node has been created?"""
         self.nrowsinbuf = None
@@ -256,10 +252,12 @@ class Leaf(Node):
             self.byteorder = byteorder
             """The byte ordering of the leaf data *on disk*."""
 
+        self._want_track_times = track_times
+
         # Existing filters need not be read since `filters`
         # is a lazy property that automatically handles their loading.
 
-        super(Leaf, self).__init__(parentnode, name, _log)
+        super().__init__(parentnode, name, _log)
 
     def __len__(self):
         """Return the length of the main dimension of the leaf data.
@@ -276,24 +274,18 @@ class Leaf(Node):
         """The string representation for this object is its pathname in the
         HDF5 object tree plus some additional metainfo."""
 
-        # Get this class name
-        classname = self.__class__.__name__
-        # The title
-        title = self._v_title
-        # The filters
-        filters = ""
+        filters = []
         if self.filters.fletcher32:
-            filters += ", fletcher32"
+            filters.append("fletcher32")
         if self.filters.complevel:
             if self.filters.shuffle:
-                filters += ", shuffle"
-            filters += ", %s(%s)" % (self.filters.complib,
-                                     self.filters.complevel)
-        return "%s (%s%s%s) %r" % \
-               (self._v_pathname, classname, self.shape, filters, title)
+                filters.append("shuffle")
+            if self.filters.bitshuffle:
+                filters.append("bitshuffle")
+            filters.append(f"{self.filters.complib}({self.filters.complevel})")
+        return (f"{self._v_pathname} ({self.__class__.__name__}"
+                f"{self.shape}{', '.join(filters)}) {self._v_title!r}")
 
-    # Private methods
-    # ~~~~~~~~~~~~~~~
     def _g_post_init_hook(self):
         """Code to be run after node creation and before creation logging.
 
@@ -301,7 +293,7 @@ class Leaf(Node):
 
         """
 
-        super(Leaf, self)._g_post_init_hook()
+        super()._g_post_init_hook()
         if self._v_new:  # set flavor of new node
             if self._flavor is None:
                 self._flavor = internal_flavor
@@ -315,8 +307,6 @@ class Leaf(Node):
             else:
                 self._flavor = internal_flavor
 
-    _g_postInitHook = previous_api(_g_post_init_hook)
-
     def _calc_chunkshape(self, expectedrows, rowsize, itemsize):
         """Calculate the shape for the HDF5 chunk."""
 
@@ -328,6 +318,38 @@ class Leaf(Node):
         MB = 1024 * 1024
         expected_mb = (expectedrows * rowsize) // MB
         chunksize = calc_chunksize(expected_mb)
+        complib = self.filters.complib
+        if (complib is not None and
+            complib.startswith("blosc2") and
+            self._c_classid == 'TABLE'):
+            # Blosc2 can introspect into blocks, so we can increase the
+            # chunksize for improving HDF5 perf for its internal btree.
+            # For the time being, this has been implemented efficiently
+            # just for tables, but in the future *Array objects could also
+            # be included.
+            # In Blosc2, the role of HDF5 chunksize could played by the
+            # Blosc2 blocksize.
+            self._v_blocksize = chunksize
+            # Use a decent default value for chunksize
+            chunksize *= 16
+            # Now, go explore the L3 size and try to find a smarter chunksize
+            cpu_info = cpuinfo.get_cpu_info()
+            if 'l3_cache_size' in cpu_info:
+                # In general, is a good idea to set the chunksize equal to L3
+                l3_cache_size = cpu_info['l3_cache_size']
+                # cpuinfo sometimes returns cache sizes as strings (like,
+                # "4096 KB"), so refuse the temptation to guess and use the
+                # value only when it is an actual int.
+                # Also, sometimes cpuinfo does not return a correct L3 size;
+                # so in general, enforcing L3 > L2 is a good sanity check.
+                l2_cache_size = cpu_info.get('l2_cache_size', "Not found")
+                if (type(l3_cache_size) is int and
+                    type(l2_cache_size) is int and
+                    l3_cache_size > l2_cache_size):
+                    chunksize = l3_cache_size
+            # In Blosc2, the chunksize cannot be larger than 2 GB - BLOSC2_MAX_BUFFERSIZE
+            if chunksize > 2**31 - 32:
+                chunksize = 2**31 - 32
 
         maindim = self.maindim
         # Compute the chunknitems
@@ -338,15 +360,15 @@ class Leaf(Node):
         chunkshape = list(self.shape)
         # Check whether trimming the main dimension is enough
         chunkshape[maindim] = 1
-        newchunknitems = numpy.prod(chunkshape, dtype=SizeType)
+        newchunknitems = np.prod(chunkshape, dtype=SizeType)
         if newchunknitems <= chunknitems:
             chunkshape[maindim] = chunknitems // newchunknitems
         else:
             # No, so start trimming other dimensions as well
-            for j in xrange(len(chunkshape)):
+            for j in range(len(chunkshape)):
                 # Check whether trimming this dimension is enough
                 chunkshape[j] = 1
-                newchunknitems = numpy.prod(chunkshape, dtype=SizeType)
+                newchunknitems = np.prod(chunkshape, dtype=SizeType)
                 if newchunknitems <= chunknitems:
                     chunkshape[j] = chunknitems // newchunknitems
                     break
@@ -366,22 +388,9 @@ class Leaf(Node):
         buffersize = params['IO_BUFFER_SIZE']
         if rowsize != 0:
             nrowsinbuf = buffersize // rowsize
-            if self.__class__.__name__ == "Table":
-                # The number of rows in buffer needs to be an exact multiple of
-                # chunkshape[0] for queries using indexed columns.
-                # Fixes #319 and probably #409 too.
-                nrowsinbuf -= nrowsinbuf % self.chunkshape[0]
         else:
             nrowsinbuf = 1
 
-        # tableextension.pyx performs an assertion
-        # to make sure nrowsinbuf is greater than or
-        # equal to the chunksize.
-        # See gh-206 and gh-238
-        if self.chunkshape is not None and self.__class__.__name__ == "Table":
-            if nrowsinbuf < self.chunkshape[0]:
-                nrowsinbuf = self.chunkshape[0]
-
         # Safeguard against row sizes being extremely large
         if nrowsinbuf == 0:
             nrowsinbuf = 1
@@ -409,13 +418,11 @@ very small/large chunksize, you may want to increase/decrease it."""
         if warn_negstep and step and step < 0:
             raise ValueError("slice step cannot be negative")
 
-        #if start is not None: start = long(start)
-        #if stop is not None: stop = long(stop)
-        #if step is not None: step = long(step)
+        # if start is not None: start = long(start)
+        # if stop is not None: stop = long(stop)
+        # if step is not None: step = long(step)
 
-        return slice(start, stop, step).indices(long(nrows))
-
-    _processRange = previous_api(_process_range)
+        return slice(start, stop, step).indices(int(nrows))
 
     # This method is appropriate for calls to read() methods
     def _process_range_read(self, start, stop, step, warn_negstep=True):
@@ -423,7 +430,7 @@ very small/large chunksize, you may want to increase/decrease it."""
         if start is not None and stop is None and step is None:
             # Protection against start greater than available records
             # nrows == 0 is a special case for empty objects
-            if nrows > 0 and start >= nrows:
+            if 0 < nrows <= start:
                 raise IndexError("start of range (%s) is greater than "
                                  "number of rows (%s)" % (start, nrows))
             step = 1
@@ -436,8 +443,6 @@ very small/large chunksize, you may want to increase/decrease it."""
                                                 warn_negstep=warn_negstep)
         return (start, stop, step)
 
-    _processRangeRead = previous_api(_process_range_read)
-
     def _g_copy(self, newparent, newname, recursive, _log=True, **kwargs):
         # Compute default arguments.
         start = kwargs.pop('start', None)
@@ -480,7 +485,7 @@ very small/large chunksize, you may want to increase/decrease it."""
         return new_node
 
     def _g_fix_byteorder_data(self, data, dbyteorder):
-        "Fix the byteorder of data passed in constructors."
+        """Fix the byteorder of data passed in constructors."""
         dbyteorder = byteorders[dbyteorder]
         # If self.byteorder has not been passed as an argument of
         # the constructor, then set it to the same value of data.
@@ -527,30 +532,30 @@ very small/large chunksize, you may want to increase/decrease it."""
         point-wise selection.
 
         """
-
+        input_key = key
         if type(key) in (list, tuple):
             if isinstance(key, tuple) and len(key) > len(self.shape):
-                raise IndexError("Invalid index or slice: %r" % (key,))
+                raise IndexError(f"Invalid index or slice: {key!r}")
             # Try to convert key to a numpy array.  If not possible,
             # a TypeError will be issued (to be catched later on).
             try:
-                key = numpy.array(key)
+                key = toarray(key)
             except ValueError:
-                raise TypeError("Invalid index or slice: %r" % (key,))
-        elif not isinstance(key, numpy.ndarray):
-            raise TypeError("Invalid index or slice: %r" % (key,))
+                raise TypeError(f"Invalid index or slice: {key!r}")
+        elif not isinstance(key, np.ndarray):
+            raise TypeError(f"Invalid index or slice: {key!r}")
 
         # Protection against empty keys
         if len(key) == 0:
-            return numpy.array([], dtype="i8")
+            return np.array([], dtype="i8")
 
         if key.dtype.kind == 'b':
             if not key.shape == self.shape:
                 raise IndexError(
                     "Boolean indexing array has incompatible shape")
             # Get the True coordinates (64-bit indices!)
-            coords = numpy.asarray(key.nonzero(), dtype='i8')
-            coords = numpy.transpose(coords)
+            coords = np.asarray(key.nonzero(), dtype='i8')
+            coords = np.transpose(coords)
         elif key.dtype.kind == 'i' or key.dtype.kind == 'u':
             if len(key.shape) > 2:
                 raise IndexError(
@@ -559,18 +564,23 @@ very small/large chunksize, you may want to increase/decrease it."""
                 if key.shape[0] != len(self.shape):
                     raise IndexError(
                         "Coordinate indexing array has incompatible shape")
-                coords = numpy.asarray(key, dtype="i8")
-                coords = numpy.transpose(coords)
+                coords = np.asarray(key, dtype="i8")
+                coords = np.transpose(coords)
             else:
                 # For 1-dimensional datasets
-                coords = numpy.asarray(key, dtype="i8")
+                coords = np.asarray(key, dtype="i8")
 
             # handle negative indices
+            base = coords if coords.base is None else coords.base
+            if base is input_key:
+                # never modify the original "key" data
+                coords = coords.copy()
+
             idx = coords < 0
             coords[idx] = (coords + self.shape)[idx]
 
             # bounds check
-            if numpy.any(coords < 0) or numpy.any(coords >= self.shape):
+            if np.any(coords < 0) or np.any(coords >= self.shape):
                 raise IndexError("Index out of bounds")
         else:
             raise TypeError("Only integer coordinates allowed.")
@@ -579,12 +589,7 @@ very small/large chunksize, you may want to increase/decrease it."""
             coords = coords.copy()
         return coords
 
-    _pointSelection = previous_api(_point_selection)
-
-    # Public methods
-    # ~~~~~~~~~~~~~~
     # Tree manipulation
-    # `````````````````
     def remove(self):
         """Remove this node from the hierarchy.
 
@@ -692,10 +697,7 @@ very small/large chunksize, you may want to increase/decrease it."""
 
         return self._f_isvisible()
 
-    isVisible = previous_api(isvisible)
-
     # Attribute handling
-    # ``````````````````
     def get_attr(self, name):
         """Get a PyTables attribute from this node.
 
@@ -705,8 +707,6 @@ very small/large chunksize, you may want to increase/decrease it."""
 
         return self._f_getattr(name)
 
-    getAttr = previous_api(get_attr)
-
     def set_attr(self, name, value):
         """Set a PyTables attribute for this node.
 
@@ -716,8 +716,6 @@ very small/large chunksize, you may want to increase/decrease it."""
 
         self._f_setattr(name, value)
 
-    setAttr = previous_api(set_attr)
-
     def del_attr(self, name):
         """Delete a PyTables attribute from this node.
 
@@ -727,10 +725,7 @@ very small/large chunksize, you may want to increase/decrease it."""
 
         self._f_delattr(name)
 
-    delAttr = previous_api(del_attr)
-
     # Data handling
-    # `````````````
     def flush(self):
         """Flush pending data to disk.
 
@@ -766,7 +761,7 @@ very small/large chunksize, you may want to increase/decrease it."""
         self._g_close()
 
         # Close myself as a node.
-        super(Leaf, self)._f_close()
+        super()._f_close()
 
     def close(self, flush=True):
         """Close this node in the tree.
@@ -776,11 +771,3 @@ very small/large chunksize, you may want to increase/decrease it."""
         """
 
         self._f_close(flush)
-
-
-## Local Variables:
-## mode: python
-## py-indent-offset: 4
-## tab-width: 4
-## fill-column: 72
-## End:
